#!/usr/bin/env python

import re, sys
from decimal import *
from session import Session, Hand
from pyparsing import Forward, Regex

# The GenericParser is subclassed by specialised parsers for histories from different sites
# When instantiated it imports the rest of the module

class Dispatcher(object):
    newlines = re.compile("\n\n+")
    def __init__(self):
        self.handType = Forward()
        self.handTypes = {}
        self.parsers = {}

    def register(self,parser,regex):
        self.handTypes[parser.__class__.__name__] = re.compile(regex)
        self.parsers[parser.__class__.__name__] = parser
#        self.handType = re.compile( "|".join( map(re.escape,  self.handTypes.values() ) ) )  # dont need this
#        self.handType << Regex("|".join( self.handTypes.values() ) )
        
    def dispatch(self, hand):
        """\
Dispatches to correct parser based on site / gametype
"""
        for (k, v) in self.handTypes.items():
            if v.match(hand):
                parserobject = self.parsers[k]
                print "Parsed by: " + k
                parserobject.parser().parseString(hand)

    def parseSession(self, sessionText):
        # take a history and parse each hand individually, appending them them into a list
        # parseSession differs from parseHands in that it is assumed that the hands are a contigious single table session
        self.hands = []
        handsList = re.sub('\r','', sessionText) # unDOSify
        handsList = handsList.strip()
        handsList = re.split(Dispatcher.newlines, handsList)
        for handText in handsList:
            self.dispatch(handText)

class GenericParser(object):
    """\
parseHand : parse a single hand and return a Hand
parseSession : parse a lot of hands, all from the same table, and return a Session

Perhaps this decoupling is not necessary. After all we could simple make
a __str__ method for Session and Hand classes and print the whole session after
it has been built.
"""



    def __init__(self, interface=None):
        """\
interface - a class that does something with the extracted information
    PrettyPrintInterface (default) : print everything to stdout
    HandBuilderInterface :  build a Hand object at self.currentHand
"""

        if interface is not None:
            self.interface = interface
        else:
            self.interface = PrettyPrintInterface()

        self.PLAYER = Forward()
        self.session = None



    ###
    #
    # Generic ParseActions
    #
    # Generally, we call the relevant interface function.
    #
    # In the subclass these methods MUST be extended: overridden and then called with the correct parameters
    # (I'll certainly consider suggestions for a better architecture - mct)
    # Otherwise the parser could be calling them directly with tokens and whatnot and that won't do.
    #
    # In addition, we maintain the list of players and the PLAYER rule.
    #
 
    def newHand(self, handNumber, startTime, smallblind,  bigblind):
        self.players = []   # reinitialise to empty.
        self.interface.newHand(handNumber, startTime,  smallblind,  bigblind)

    def addPlayer(self, seat=None, name=None, chips=None):
        """parseAction to maintain the list of players and the PLAYER rule"""
        if name and chips and seat:
            self.interface.addPlayer(seat, name, chips)
            self.players.append(name)
            self.PLAYER << Regex("|".join(map(re.escape, self.players)))

    def setButton(self, buttonSeat):
        self.interface.setButton(buttonSeat)

    def setHoleCards(self, name, hole):
        """name : a player, known to have matched the PLAYER rule
        hole : two cards"""
        self.interface.setHoleCards(name, hole)

    def setFlop(self, flopCards):
        self.interface.setFlop(flopCards)

    def setTurn(self, turnCard):
        self.interface.setTurn(turnCard)

    def setRiver(self, riverCard):
        self.interface.setRiver(riverCard)

    def addBlind(self, name, amount):
        self.interface.addBlind(name, amount)

    def addCall(self, name, amount):
        self.interface.addCall(name, amount)

    def addCheck(self, name):
        self.interface.addCheck(name)

    def addBet(self, name, amount):
        self.interface.addBet(name, amount)

    def addRaise(self, name, amount):
        """amount : the amount raised *to*"""
        self.interface.addRaise(name, amount)

    def addFold(self, name):
        self.interface.addFold(name)

    def addUncalled(self, name, amount):
        self.interface.addUncalled(name, amount)

    def collectedAmount(self, name, amount):
        """Important : call only once per winner! Sometimes it is reported more than once."""
        # todo: sanity check uniqueness
        self.interface.collectedAmount(name, amount)

    def setRake(self, amount):
        self.interface.setRake(amount)

    def endHand(self):
        self.interface.endHand()

    def addMuckedCards(self, name, hole):
        ### do we care if cards are flashed, seen at showdown, or what? they are different ways of acquiring the information and incorporate different biases. If a very good player is flashing cards, it is obviously misinformation. If they are seen at showdown it is less likely so."""
        self.interface.setHoleCards(name, hole)



    def parser(self):
        """\
This is overridden by subclasses."""




class PrettyPrintInterface(object):
    """\
Just prints everything.

Translates all amounts to big blinds for ease of comprehension.

Prints all observed hole cards for each hand.

It may be of interest that unicodes for spade, heart, diamond, club are (according to wikipedia):
    U+2660 dec: 9824    U+2661 dec: 9825    U+2662 dec: 9826    U+2663 dec: 9827
"""
    conversion = {
         's' : u'\u2660',
         'h' : u'\u2661', 
         'd' : u'\u2662', 
         'c' : u'\u2663'
         }

    def tounicodesuit(self, s):
        retval = u""
        for c in s:
            try:
                c = PrettyPrintInterface.conversion[c]
            except KeyError:
                pass
            retval += c
        return retval

    currentHand = None

    def newHand(self, handNumber, startTime,  smallblind,  bigblind):
        self.sb = smallblind
        self.bb = bigblind
        print "Hand", handNumber, "at", startTime

    def addPlayer(self, seat=None, name=None, chips=None):
        print "Seat", seat, ":", name, "(", chips/self.bb, ")"

    def setButton(self, buttonSeat):
        print "Seat", buttonSeat, "is the button"

    def addBlind(self, name, amount):
        print "%s pays blind %.1f" % (name,  (amount/self.bb), )

    def setHoleCards(self, name, hole):
#        hole = [self.tounicodesuit(h) for h in hole]
        print "** Hole cards **", hole, name

    def setFlop(self, flopCards):
        print "** Flop ** %s" % (flopCards, )

    def setTurn(self, turnCard):
        print "** Turn ** [%s]" % (turnCard, )

    def setRiver(self, riverCard):
        print "** River** [%s]" % (riverCard, )

    def addCall(self, name, amount):
        print "%s calls" % (name, )

    def addCheck(self, name):
        print "%s checks" %  (name, )

    def addBet(self, name, amount):
        print "%s bets %.1f" % (name,  amount/self.bb,  )

    def addRaise(self, name, amount):
        print "%s raises to %.1f" % ( name, amount/self.bb,  )

    def addFold(self, name):
        print name, "folds"

    def addUncalled(self, name, amount):
        print "uncalled bet (%.1f) returned to %s" % ( (amount/self.bb), name , )

    def collectedAmount(self, name, amount):
        print "%s collected %.1f" % (name,  (amount/self.bb) , ) # note collected != net won

    def setRake(self, amount):
        print "rake: %.1f" % (amount/self.bb)

    def endHand(self):
        print

    def addMuckedCards(self, name, hole):
#        hole = [self.tounicodesuit(h) for h in hole]
        print name, "mucked", hole
    

class HandBuilderInterface(object):
    """Builds a Hand object"""
    def newHand(self, handNumber, startTime):
        self.currentHand = Hand(handNumber, startTime)

    def addPlayer(self, name=None, chips=None, seat=None):
        """parseAction to maintain the list of players and the PLAYER rule"""
        if name and chips and seat:
            self.currentHand.addPlayer(seat, name, chips)

    def setButton(self, buttonSeat):
        self.currentHand.setButton(buttonSeat)

    def setHoleCards(self, name, hole):
        """name : a player, known to have matched the PLAYER rule
        hole : two cards"""
        self.currentHand.setHoleCards(name, hole)

    def setFlop(self, flopCards):
        self.currentHand.setFlop(flopCards)

    def setTurn(self, turnCard):
        self.currentHand.setTurn(turnCard)

    def setRiver(self, riverCard):
        self.currentHand.setRiver(riverCard)

    def addBlind(self, name, amount):
        self.currentHand.addBlind(name, amount)

    def addCall(self, name, amount):
        self.currentHand.addCall(name, amount)

    def addCheck(self, name):
        self.currentHand.addCall(name, amount=0)

    def addBet(self, name, amount):
        self.currentHand.addBet(name, amount)

    def addRaise(self, name, amount):
        """amount : the amount raised *to*"""
        self.currentHand.addRaise(name, amount)

    def addFold(self, name):
        self.currentHand.addFold(name)

    def uncalledBet(self, name, amount):
        self.currentHand.addUncalled(name, amount)

    def collectedAmount(self, name, amount):
        """Important : call only once per winner! Sometimes it is reported more than once."""
        # todo: sanity check uniqueness
        self.currentHand.addWinner(name, amount)

    def setRake(self, amount):
        self.currentHand.setRake(amount)

    def endHand(self):
        self.currentHand.endHand()

    def addMuckedCards(self, name, hole):
        ### do we care if cards are flashed, seen at showdown, or what? they are different ways of acquiring the information and incorporate different biases. If a very good player is flashing cards, it is obviously misinformation. If they are seen at showdown it is less likely so."""
        self.currentHand.setHoleCards(name, hole)

#class GenericParserInterfacePyPokeHandBuilder(GenericParserInterfaceHandBuilder):
#    """Adds signals to report progress to the pyqt-based program"""
#    pass



def main():
    dispatcher = Dispatcher()
    # going to make a parser here but really this should be done .. when? where?
    # by passing self, the init of PSNLHECGP should register with this Dispatcher.
    from PokerStarsNLHECashGame import PokerStarsNLHECashGameParser
    parser = PokerStarsNLHECashGameParser(dispatcher)
    
    fname = "PokerStarsNLHECashGame.sample.txt"
    fh = None
    try:
        fh = open(fname)
        history = fh.read()
    finally:
        if fh is not None:
            fh.close()
    
    dispatcher.parseSession(history)
    
if __name__ == '__main__':
    main()
    
